"""
@author: Federica Colombo
         Psychiatry and Clinical Psychobiology Unit, Division of Neuroscience, 
         IRCCS San Raffaele Scientific Institute, Milan, Italy
"""

import numpy as np
from scipy.optimize import linear_sum_assignment
from sklearn.metrics import matthews_corrcoef, accuracy_score, f1_score, \
    precision_score, recall_score

def kuhn_munkres_algorithm(true_lab, pred_lab):
    """
    Function that implements the Kuhn-Munkres algorithm method.
    It selects the best label permutation of the
    predicted labels that minimizes the
    misclassification error when compared to the true labels.
    In order to allow for the investigation
    of replicability of findings between training and test sets,
    in the context of reval we permute
    clustering labels to match classification labels,
    in order to retain the label organization
    based on training dataset. This because otherwise we would loose the
    correspondence between training and test
    sets labels.

    :param true_lab: classification algorithm labels (for reval).
    :type true_lab: ndarray, (n_samples,)
    :param pred_lab: clustering algorithm labels (for reval).
    :type pred_lab: ndarray, (n_samples,)
    :return: permuted labels that minimize the misclassification error.
    :rtype: ndarray, (n_samples,)
    """
    if not (isinstance(true_lab, np.ndarray) and isinstance(pred_lab, np.ndarray)):
        true_lab, pred_lab = np.array(true_lab), np.array(pred_lab)
    if (true_lab.dtype == 'int' or true_lab.dtype == 'int32') and (
            pred_lab.dtype == 'int' or pred_lab.dtype == 'int32'):
        wmat = _build_weight_mat(true_lab, pred_lab)
        new_pred_lab = list(linear_sum_assignment(wmat)[1])
        try:
            pred_perm = np.array([new_pred_lab.index(i) for i in pred_lab])
            return pred_perm
        # Catch the case in which predicted array has more unique labels than true array. Raise ValueError
        # if input has length 1.
        except ValueError:
            if len(true_lab) == 1 or len(pred_lab) == 1:
                raise ValueError("Dimensions of input array should be greater than 1.")
            # True labels are less than predicted labels. Permuting only the available labels.
            else:
                pred_perm = np.array([], dtype=int)
                for i in pred_lab:
                    if len(new_pred_lab) <= i or i < 0:
                        pred_perm = np.append(pred_perm, i)
                    else:
                        pred_perm = np.append(pred_perm, new_pred_lab.index(i))
                return pred_perm
    else:
        raise TypeError(f'Input variables should be array-like structures of integers.'
                        f' Received ({true_lab.dtype}, {pred_lab.dtype})')


def _build_weight_mat(true_lab, pred_lab):
    """
    This private function allows to build a weight matrix to select the best permutation of
    predicted labels that minimizes the misclassification error when compared to true labels.

    :param true_lab: clustering labels, (n_samples,).
    :type true_lab: ndarray
    :param pred_lab: classification labels, (n_samples,).
    :type pred_lab: ndarray
    :return: weight matrix (number of true classes, number of true classes).
    :rtype: ndarray
    """
    if not (isinstance(true_lab, np.ndarray) and isinstance(pred_lab, np.ndarray)):
        raise TypeError("Numpy array required.")
    nclass, nobs = len(set(true_lab)), len(true_lab)
    wmat = np.zeros((nclass, nclass))
    for lab in range(nclass):
        for plab in range(lab, nclass):
            n_intersec = len(set(np.flatnonzero(true_lab == lab)).intersection(
                set(np.flatnonzero(pred_lab == plab))))
            w = (nobs - n_intersec) / nobs
            if lab == plab:
                wmat[lab, plab] = w
            else:
                wmat[lab, plab] = w
                n_intersec = len(set(np.flatnonzero(true_lab == plab)).intersection(
                    set(np.flatnonzero(pred_lab == lab))))
                w = (nobs - n_intersec) / nobs
                wmat[plab, lab] = w
    return wmat


def compute_metrics(class_labels, clust_labels, perm=False):
    """
    Function that computes useful classification metrics. If needed the
    clustering labels are permuted with :reval.utils.kuhn_munkres_algorithm:
    The function returns a dictionary with ACC, MCC, F1, precision, recall as keys
    for accuracy, Matthews correlation coefficient, F1 score, precision, and recall,
    respectively.

    :param class_labels: labels returned by the classifier.
    :type class_labels: array-like
    :param clust_labels: labels returned by the clustering.
    :type clust_labels: array-like
    :param perm: flag to enable permutation of clustering labels, default False.
    :type perm: bool
    :return: dictionary of scores.
    :rtype: dict
    """
    if perm:
        perm_clust_labels = kuhn_munkres_algorithm(class_labels, clust_labels)
    else:
        perm_clust_labels = clust_labels
    if len(set(class_labels)) == 2 and len(set(perm_clust_labels)) == 2:
        scores = {'ACC': accuracy_score(class_labels, perm_clust_labels),
                  'MCC': matthews_corrcoef(class_labels, perm_clust_labels),
                  'F1': f1_score(class_labels, perm_clust_labels),
                  'precision': precision_score(class_labels, perm_clust_labels),
                  'recall': recall_score(class_labels, perm_clust_labels)}
    else:
        scores = {'ACC': accuracy_score(class_labels, perm_clust_labels),
                  'MCC': matthews_corrcoef(class_labels, perm_clust_labels),
                  'F1': f1_score(class_labels, perm_clust_labels, average='macro'),
                  'precision': precision_score(class_labels, perm_clust_labels,
                                               average='macro', zero_division=0),
                  'recall': recall_score(class_labels, perm_clust_labels, average='macro',
                                         zero_division=0)}
    return scores





"""
@author: Federica Colombo
         Psychiatry and Clinical Psychobiology Unit, Division of Neuroscience, 
         IRCCS San Raffaele Scientific Institute, Milan, Italy
"""

import numpy as np
from sklearn.preprocessing import StandardScaler
import math

def select_best(data, modalities, covariates, c, int_measure, select='max', nclust_range=None):
    """
    Select the best number of clusters that minimizes/maximizes
    the internal measure selected.

    :param data: dataset.
    :type data: array-like
    :param c: clustering algorithm class.
    :type c: obj
    :param int_measure: internal measure function.
    :type int_measure: obj
    :param preprocessing: data reduction algorithm class, default None.
    :type preprocessing: obj
    :param select: it can be 'min', if the internal measure is to be minimized
        or 'max' if the internal measure should be macimized.
    :type select: str
    :param nclust_range: Range of clusters to consider, default None.
    :type nclust_range: list
    :param combined_data: define whether multimodal data are used as input features. 
        If True, different sets of covariates will be applied for each modality
        e.g. correction for TIV only for grey matter features. Default False
    :type combined_data: boolean value
    :return: internal score and best number of clusters.
    :rtype: float, int
    """
    X_dic = {mod:None for mod in modalities}
    cov_dic = {mod:None for mod in modalities}
    
    for mod in modalities:
       X_dic[mod] = np.array(modalities[mod])
       cov_dic[mod] = np.array(covariates[mod])
       
    # normalize the covariate z-scoring
    X_scaled_dic = {mod:None for mod in modalities}
    cov_scaled_dic = {mod:None for mod in modalities}
    scaler = StandardScaler()
   
    for mod in modalities:
       X_scaled_dic[mod] = scaler.fit_transform(X_dic[mod])
       cov_scaled_dic[mod]= scaler.fit_transform(cov_dic[mod])
   
   
    # Adjust data for confounds of covariate
    X_cor_dic = {mod:None for mod in modalities}
    
    for mod in modalities:
       beta = np.linalg.lstsq(cov_scaled_dic[mod], X_scaled_dic[mod], rcond=None)
       X_cor_dic[mod] = (X_scaled_dic[mod].T - beta[0].T @ cov_scaled_dic[mod].T).T
    
    #mv_data = np.concatenate([X_cor_dic[mod]for mod in X_cor_dic], axis=1)
    mv_data = [np.array(X_cor_dic[mod]) for mod in X_cor_dic]
        
    if nclust_range is not None:
        scores = []
        label_vect = []
        for ncl in nclust_range:
            if 'n_clusters' in c.get_params().keys():
                c.n_clusters = ncl
            else:
                c.n_components = ncl
            label = c.fit_predict(mv_data)
            mv_embedding = c.embedding_
            scores.append(int_measure(mv_embedding, label))
            label_vect.append(label)
    else:
        label = c.fit_predict(mv_data)
        mv_embedding = c.embedding_
        scores.append(int_measure(mv_embedding, label))
        return best, len([lab for lab in np.unique(label) if lab >= 0]), label
        
    if select == 'max':
        best = np.where(np.array(scores) == max(scores))[0]
    elif select == 'min':
        best = np.where(np.array(scores) == min(scores))[0]
    if len(set(label_vect[int(max(best))])) == nclust_range[int(max(best))]:
        return scores[int(max(best))], nclust_range[int(max(best))], label_vect[int(max(best))]
    else:
        return scores[int(max(best))], len(set(label_vect[int(max(best))])), label_vect[int(max(best))]


def evaluate_best(data, modalities, covariates, c, int_measure, ncl=None):
    """
    Function that, given a number of clusters, returns the corresponding internal measure
    for a dataset.

    :param data: dataset.
    :type data: array-like
    :param c: clustering algorithm class.
    :type c: obj
    :param int_measure: internal measure function.
    :type int_measure: obj
    :param preprocessing:  dimensionality reduction algorithm class, default None.
    :type preprocessing: obj
    :param ncl: number of clusters.
    :type ncl: int
    :param combined_data: define whether multimodal data are used as input features. 
        If True, different sets of covariates will be applied for each modality
        e.g. correction for TIV only for grey matter features. Default False
    :type combined_data: boolean value
    :return: internal score.
    :rtype: float
    """
    X_dic = {mod:None for mod in modalities}
    cov_dic = {mod:None for mod in modalities}
    
    for mod in modalities:
       X_dic[mod] = np.array(modalities[mod])
       cov_dic[mod] = np.array(covariates[mod])
       
    # normalize the covariate z-scoring
    X_scaled_dic = {mod:None for mod in modalities}
    cov_scaled_dic = {mod:None for mod in modalities}
    scaler = StandardScaler()
   
    for mod in modalities:
       X_scaled_dic[mod] = scaler.fit_transform(X_dic[mod])
       cov_scaled_dic[mod]= scaler.fit_transform(cov_dic[mod])
   
   
    # Adjust data for confounds of covariate
    X_cor_dic = {mod:None for mod in modalities}
    
    for mod in modalities:
       beta = np.linalg.lstsq(cov_scaled_dic[mod], X_scaled_dic[mod], rcond=None)
       X_cor_dic[mod] = (X_scaled_dic[mod].T - beta[0].T @ cov_scaled_dic[mod].T).T

    mv_data = np.concatenate([X_cor_dic[mod]for mod in X_cor_dic], axis=1)
    
    if 'n_clusters' in c.get_params().keys():
        c.n_clusters = ncl
    else:
        c.n_components = ncl
        label = c.fit_predict(mv_data)
    
    return int_measure(mv_data, label)



"""
@author: Federica Colombo
         Psychiatry and Clinical Psychobiology Unit, Division of Neuroscience, 
         IRCCS San Raffaele Scientific Institute, Milan, Italy
"""

from sklearn.model_selection import ParameterGrid
from mvneureval.best_nclust_cv_multiview import FindBestClustCVMultiview
from mvneureval.relative_validation_multiview import RelativeValidationMultiview
import multiprocessing as mp
import logging
import numpy as np
import itertools

from sklearn.cluster import AgglomerativeClustering

logging.basicConfig(format='%(asctime)s, %(levelname)s %(message)s',
                    datefmt='%H:%M:%S',
                    level=logging.INFO)


class ParamSelectionMultiview(RelativeValidationMultiview):
    """
    Class that implements grid search cross-validation in parallel to select
    the best combinations of parameters for fixed classifier/clustering algorithms.
    If a preprocessing method (e.g. dimensionality reduction algorithm) is specified,
    also the specified parameters will be selected through grid-search cross-validation.

    :param params: dictionary of dictionaries of the form {'s': {classifier parameter grid},
        'c': {clustering parameter grid}, 'preprocessing': {preprocessing parameter grid}}. If one of the two dictionary of parameters is not
        available, initialize key but leave dictionary empty.
    :type params: dict
    :param cv: cross-validation folds.
    :type cv: int
    :param clust_range: list with number of clusters to investigate.
    :type clust_range: list
    :param n_jobs: number of jobs to run in parallel, default (number of cpus - 1).
    :type n_jobs: int
    :param iter_cv: number of repeated cv loops, default 1.
    :type iter_cv: int
    :param strat: stratification vector for cross-validation splits, default None.
    :type strat: numpy array
    :param combined_data: if True, only grey matter features will be adjusted for TIV.
        Otherwise, both grey and white matter features will be adjusted for the same set of covariates.
    :type combined_data: boolean value

    :attribute: cv_results_ cross-validation results that can be directly transformed to
        a dataframe. Key names: classifier parameters, clustering parameters,
        'best_nclust', 'mean_train_score', 'sd_train_score',
        'mean_val_score', 'sd_val_score', 'validation_meanerror'. Dictionary of lists.
    :attribute: best_param_ best solution(s) selected (minimum validation error). List.
    :attribute: best_index_ index/indices of the best solution(s). Values correspond to the
        rows of the `cv_results_` table. List.
    """

    def __init__(self, params, cv, c, s, nrand,
                n_jobs=1, iter_cv=1, strat=None, clust_range=None):
        super().__init__(c, s, nrand)
        self.params = params
        self.cv = cv
        self.iter_cv = iter_cv
        self.clust_range = clust_range
        if abs(n_jobs) > mp.cpu_count():
            self.n_jobs = mp.cpu_count()
        else:
            self.n_jobs = abs(n_jobs)
        self.strat = strat

    def fit(self, data_tr, mod_tr, cov_tr, nclass=None):
        """
        Class method that performs grid search cross-validation on training data. It
        deals with the error due to wrong parameter combinations (e.g., ward linkage
        with no euclidean affinity). If the true number of classes is know, the method
        selects both the best parameter combination that selects the true number of clusters
        (minimum stability) and the best parameter combination that minimizes
        overall stability.

        :param data_tr: training dataset.
        :type data_tr: numpy array
        :param cov_tr: covariates dataset.
        :type cov_tr: numpy array
        :param nclass: number of true classes, default None.
        :type nclass: int
        """

        grid = {'c': ParameterGrid(self.params['c']), 's': ParameterGrid(self.params['s'])}
        new_grid = list(itertools.product( grid['c'], grid['s']))
        new_params = [(data_tr, mod_tr, cov_tr, ng[0], ng[1]) for ng in new_grid]
        
        if len(new_grid) != len(new_params):
            logging.info(f"Dropped {len(new_grid) - len(new_params)} out of {len(new_grid)} parameter "
                          f"combinations "
                          f"due to {self.clust_method} class requirements.")

        logging.info(f'Running {len(new_params)} combinations of '
                      f'parameters...\n')

        p = mp.Pool(processes=self.n_jobs)
        
        out = list(zip(*p.starmap(self._run_gridsearchcv, new_params))) #new_params
        p.close()
        p.join()
    
        # cv_results_
        res_dict = _create_result_table(out)
        ParamSelectionMultiview.cv_results_ = res_dict

        # best_param_, best_index_
        val_scores = [vs for vs in res_dict['mean_val_score'] if vs is not None]
        val_idx = [idx for idx, vs in enumerate(res_dict['mean_val_score']) if vs is not None]
        if len(val_scores) > 0:
            idx_best = [val_idx[i] for i in _return_best(val_scores)]
        else:
            logging.info(f"No clustering solutions were found with any parameter combinations.")
            return self
        
    
        out_best = []
        if nclass is not None:
            logging.info(f'True number of clusters known: {nclass}\n')
            idx = np.where(np.array(res_dict['best_nclust']) == nclass)[0]
            idx_inter = set(idx).intersection(set(idx_best))
            if len(idx_inter) > 0:
                idx_best = list(idx_inter)
            else:
                if len(idx) > 0:
                    idx_true = _return_knownbest(res_dict['mean_val_score'], idx)
                    logging.info(f'Best solution(s) with true number of clusters:')
                    for bidx in idx_true:
                        for k in self.params['c'].keys():
                            logging.info(f'Parameters multiview clustering (C): {k}={res_dict[k][bidx]}')
                        for k in self.params['s'].keys():
                            logging.info(f'Parameters classifier (S): {k}={res_dict[k][bidx]}')
                        logging.info(f'Validation performance: {res_dict["validation_meanerror"][bidx]}')
                        logging.info(f'N clusters: {res_dict["best_nclust"][bidx]}\n')
                        out_best.append([res_dict[k][bidx] for k in self.params['c'].keys()] +
                                        [res_dict[k][bidx] for k in self.params['s'].keys()] +
                                        [res_dict["best_nclust"][bidx], res_dict["validation_meanerror"][bidx], res_dict["tr_label"][bidx]])
                   
        logging.info(f'Best solution(s):')
        for bidx in idx_best:
            for k in self.params['c'].keys():
                logging.info(f'Parameters multiview clustering (C): {k}={res_dict[k][bidx]}')
            for k in self.params['s'].keys():
                logging.info(f'Parameters classifier (S): {k}={res_dict[k][bidx]}')
            logging.info(f'Validation performance: {res_dict["validation_meanerror"][bidx]}')
            logging.info(f'N clusters: {res_dict["best_nclust"][bidx]}\n')
            out_best.append([res_dict[k][bidx] for k in self.params['c'].keys()] +
                            [res_dict[k][bidx] for k in self.params['s'].keys()] +
                            [res_dict["best_nclust"][bidx], res_dict["validation_meanerror"][bidx], res_dict["tr_label"][bidx]])
        
        ParamSelectionMultiview.best_param_ = out_best
        ParamSelectionMultiview.best_index_ = idx_best

        return self

    def _run_gridsearchcv(self, data, modalities, covariates, param_c, param_s):
        """
        Private method that initializes classifier/clustering/preprocessing with different
        parameter combinations and :class:`neureval.best_nclust_cv_confounds.FindBestClustCVConfounds`.

        :param data: training dataset.
        :type data: numpy array
        :param param_s: dictionary of classifier parameters.
        :type: dict
        :param param_c: dictionary of clustering parameters.
        :type param_c: dict
        :param param_preprocessing:: dictionary of preprocessing parameters.
        :type param_preprocessing: dict
        :return: performance list.
        :rtype: list
        """
        self.clust_method.set_params(**param_c)
        self.class_method.set_params(**param_s)
       
        findclust = FindBestClustCVMultiview(nfold=self.cv,
                                             c=self.clust_method,
                                             s=self.class_method,
                                             nrand=self.nrand,n_jobs=1,
                                             nclust_range=self.clust_range)
                                             
        if self.clust_range is not None:
                metric, nclbest, tr_lab = findclust.best_nclust(data, modalities, covariates, iter_cv=self.iter_cv, strat_vect=self.strat)
                #tr_lab = None
        else:
                try:
                    metric, nclbest, tr_lab = findclust.best_nclust(data, modalities, covariates, iter_cv=self.iter_cv, strat_vect=self.strat)
                except TypeError:
                    perf =  [(key, val) for key, val in param_c.items()] + \
                            [(key, val) for key, val in param_s.items()] + \
                            [('best_nclust', None),
                            ('mean_train_score', None),
                            ('sd_train_score', None),
                            ('mean_val_score', None),
                            ('sd_val_score', None),
                            ('validation_meanerror', None),
                            ('tr_label', None),
                            ]
                    return perf
                

        perf =  [(key, val) for key, val in param_c.items()] + \
                [(key, val) for key, val in param_s.items()] + \
                [('best_nclust', nclbest),
                    ('mean_train_score', np.mean(
                        findclust.cv_results_.loc[findclust.cv_results_.ncl == nclbest]['ms_tr'])),
                    ('sd_train_score', np.std(
                        findclust.cv_results_.loc[findclust.cv_results_.ncl == nclbest]['ms_tr'])),
                    ('mean_val_score', np.mean(
                        findclust.cv_results_.loc[findclust.cv_results_.ncl == nclbest]['ms_val'])),
                    ('sd_val_score', np.std(
                        findclust.cv_results_.loc[findclust.cv_results_.ncl == nclbest]['ms_val'])),
                    ('validation_meanerror', metric['val'][nclbest]),
                    ('tr_label', tr_lab),
                    ]
        return perf
    
    def _allowed_par(self, par_dict):
        """
        Private method that controls the allowed parameter combinations
        for hierarchical clustering.
    
        :param par_dict: clustering parameter grid.
        :type par_dict: dict
        :return: whether the parameter combination can be allowed.
        :rtype: bool
        """
        if isinstance(self.clust_method, AgglomerativeClustering):
            try:
                if par_dict['linkage'] == 'ward':
                    return par_dict['affinity'] == 'euclidean'
                else:
                    return True
            except KeyError:
                try:
                    return par_dict['affinity'] == 'euclidean'
                except KeyError:
                    return True
        else:
            return True



"""
Private functions
"""


def _return_best(val_scores):
    """
    Private function that returns indices corresponding to the best solution,
    i.e., those that minimize the validation stability scores.

    :param val_scores: list of validation scores averaged over cross-validation loops.
    :type val_scores: list
    :return: list of indices.
    :rtype: list
    """
    bidx = list(np.where(np.array(val_scores) == min([vs for vs in val_scores]))[0])
    return bidx


def _return_knownbest(val_perf, idx):
    """
    Private function that, given a stability score list and indices, returns the indices corresponding
    to the best solution.

    :param val_perf: list of validation scores averaged over cross-validation loops.
    :type val_perf: list
    :param idx: list of indices.
    :type idx: list
    :return: list of indices.
    :rtype: list
    """
    bidx = _return_best([val_perf[i] for i in idx])
    return [idx[b] for b in bidx]


def _create_result_table(out):
    """
    Private function that builds the performance result dictionary to be transformed to
    dataframe.

    :param out: grid search performance results.
    :type out: list
    :return: dictionary with results.
    :rtype: dict
    """
    dict_obj = {}
    for el in out:
        for key, val in el:
            if key in dict_obj:
                if not isinstance(dict_obj[key], list):
                    dict_obj[key] = [dict_obj[key]]
                dict_obj[key].append(val)
            else:
                dict_obj[key] = val
    return dict_obj



# -*- coding: utf-8 -*-
"""
@author: Federica Colombo
         Psychiatry and Clinical Psychobiology Unit, Division of Neuroscience, 
         IRCCS San Raffaele Scientific Institute, Milan, Italy
"""

from sklearn.model_selection import RepeatedKFold, RepeatedStratifiedKFold
from mvneureval.relative_validation_multiview  import RelativeValidationMultiview
from collections import namedtuple
from scipy import stats
import numpy as np
import math
import logging
import multiprocessing as mp
import itertools
import pandas as pd
import statistics

logging.basicConfig(format='%(asctime)s, %(levelname)s %(message)s',
                    datefmt='%H:%M:%S',
                    level=logging.INFO)


class FindBestClustCVMultiview(RelativeValidationMultiview):
    """
    Child class of :class:`reval.relative_validation.RelativeValidationConfounds`.
    It performs (repeated) k-fold cross validation on the training set to
    select the best number of clusters, i.e., the number that minimizes the
    normalized stability (i.e., average misclassification
    error/asymptotic misclassification rate).

    :param nfold: number of CV folds.
    :type nfold: int
    :param nclust_range: list with clusters to look for, default None.
    :type nclust_range: list of int
    :param s: classification object inherited
        from :class:`neureval.relative_validation.RelativeValidationConfounds`.
    :type s: class
    :param c: clustering object inherited
        from :class:`neureval.relative_validation.RelativeValidationConfounds`.
    :type c: class
    :param preprocessing: preprocessing method inherited
        from :class:`neureval.relative_validation.RelativeValidationConfounds`.
    :type preprocessing: class
    :param nrand: number of random labeling iterations
        to compute asymptotic misclassification rate, inherited from
        :class:`neureval.relative_validation.RelativeValidation` class.
    :type nrand: int
    :param n_jobs: number of processes to be run in parallel, default 1.
    :type n_jobs: int

    :attribute: cv_results_ dataframe with cross validation results. Columns are
        ncl = number of clusters; ms_tr = misclassification training;
        ms_val = misclassification validation.
    """

    def __init__(self, c, s, nrand, nfold=2, n_jobs=1, nclust_range=None):
        """
        Construct method.
        """
        super().__init__(c, s, nrand)
        self.nfold = nfold
        if abs(n_jobs) > mp.cpu_count():
            self.n_jobs = mp.cpu_count()
        else:
            self.n_jobs = abs(n_jobs)
        self.nclust_range = nclust_range

    def best_nclust(self, data, modalities, covariates, iter_cv=1, strat_vect=None):
        """
        This method takes as input the training dataset and the
        stratification vector (if available) and performs a
        (repeated) CV procedure to select the best number of clusters that minimizes
        normalized stability.

        :param data: training dataset.
        :type data: ndarray, (n_samples, n_features)
        :param covariates: covariates dataset.
        :type covatìriates: ndarray, (n_samples, n_covariates)
        :param iter_cv: number of iteration for repeated CV, default 1.
        :type iter_cv: integer
        :param strat_vect: vector for stratification, defaults to None.
        :type strat_vect: ndarray, (n_samples,)
        :param combined_data: if True, only grey matter features will be corrected for TIV. 
            Otherwise, both grey matter and white matter features will be adjusted for the same set of covariates.
        :type: boolean value
        :return: CV metrics for training and validation sets, best number of clusters,
            misclassification errors at each CV iteration.
        :rtype: dictionary, int, (list) if n_clusters parameter is not available
        """
        
        mv_reval = RelativeValidationMultiview(self.clust_method, self.class_method, self.nrand)

        if strat_vect is not None:
            kfold = RepeatedStratifiedKFold(n_splits=self.nfold, n_repeats=iter_cv, random_state=42)
        else:
            kfold = RepeatedKFold(n_splits=self.nfold, n_repeats=iter_cv, random_state=42)
        fold_gen = kfold.split(data, strat_vect)
        if self.nclust_range is not None:
            params = list(itertools.product([(modalities, mv_reval, covariates)],
                                            fold_gen, self.nclust_range))
        else:
            params = list(itertools.product([(modalities, mv_reval, covariates)],fold_gen))
        if self.n_jobs > 1:
            p = mp.Pool(processes=self.n_jobs)
            miscl = list(zip(*p.starmap(self._fit_multiview, params)))
            p.close()
            p.join()
            out = list(zip(*miscl))
        else:
            miscl = []
            for p in params:
                if len(p) > 2:
                    miscl.append(self._fit_multiview(data_obj=p[0],
                                                     idxs=p[1],
                                                     ncl=p[2]))
                else:
                    miscl.append(self._fit_multiview(data_obj=p[0],
                                                     idxs=p[1]))
            out = miscl

        # return dataframe attribute (cv_results_) with cv scores
        # If no point are labeled (e.g., all points assigned to -1 class by HDBSCAN)
        # the method returns
        cv_results_ = pd.DataFrame(out,
                                   columns=['ncl', 'ms_tr', 'ms_val', 'tr_labels', 'val_labels'])
        ctrl_rows = cv_results_.shape[0]
        cv_results_.dropna(axis=0, inplace=True)
        if 0 < ctrl_rows - cv_results_.shape[0] < ctrl_rows:
            logging.info("Dropped results where clustering algorithm failed to identify clusters.")
            FindBestClustCVMultiview.cv_results_ = cv_results_
        elif ctrl_rows - cv_results_.shape[0] == 0:
            FindBestClustCVMultiview.cv_results_ = cv_results_
        else:
            logging.info(f"{self.clust_method} was not able to identify any cluster. Failed run with "
                         f"{self.class_method}.")
            return None

        metrics = {'train': {}, 'val': {}}
        for ncl in cv_results_.ncl.unique():
            norm_stab_tr = cv_results_.loc[cv_results_.ncl == ncl]['ms_tr']
            norm_stab_val = cv_results_.loc[cv_results_.ncl == ncl]['ms_val']
            metrics['train'][ncl] = (np.mean(norm_stab_tr), _confint(norm_stab_tr))
            metrics['val'][ncl] = (np.mean(norm_stab_val), _confint(norm_stab_val))

        val_score = np.array([val[0] for val in metrics['val'].values()])
        bestscore = min(val_score)
        # select the cluster with the minimum misclassification error
        # and the maximum number of clusters
        if self.nclust_range is not None:
            bestncl = self.nclust_range[np.flatnonzero(val_score == bestscore)[-1]]
            # To comment
            best_idx = cv_results_.loc[cv_results_.ncl == bestncl].ms_val.idxmin()
            idx_vect = np.concatenate((params[best_idx][1][-2], params[best_idx][1][-1]))
            label_vect = np.concatenate((out[best_idx][-2], out[best_idx][-1]))
            tr_lab = [lab for _, lab in sorted(zip(idx_vect, label_vect))]
            return metrics, bestncl, tr_lab
        else:
            bestncl = list(metrics['val'].keys())[np.flatnonzero(val_score == bestscore)[-1]]
            best_idx = cv_results_.loc[cv_results_.ncl == bestncl].ms_val.idxmin()
            idx_vect = np.concatenate((params[best_idx][1][-2], params[best_idx][1][-1]))
            label_vect = np.concatenate((out[best_idx][-2], out[best_idx][-1]))
            tr_lab = [lab for _, lab in sorted(zip(idx_vect, label_vect))]
            return metrics, bestncl, tr_lab


    @staticmethod
    def _fit_multiview(data_obj, idxs, ncl=None):
        """
        Function that calls training, test, covariates, and random labeling.
    
        :param data_obj: dataset and `reval.RelativeValidationConfounds` class.
        :type data_obj: tuple
        :param cov_obj: covariates and diagnosis labels.
        :type cov_obj: tuple
        :param idxs: lists of training and validation indices.
        :type idxs: tuple
        :param ncl: number of clusters, default None
        :type ncl: int
        :return: number of clusters and misclassification errors for training and validation.
        :rtype: tuple (int, float, float)
        """
    
        modalities, mv_reval, covariates = data_obj
        tr_idx, val_idx = idxs
        mv_reval.clust_method.n_clusters = ncl
    
        # Case in which the number of clusters identified is not sufficient or points are not classified
        # (it happens particularly for SpectralClustering and HDBSCAN)
        try:
            train_cor_dic, test_cor_dic = mv_reval.GLMcorrection_multimodal(modalities, covariates, tr_idx, val_idx)
            miscl_tr, modelfit, tr_labels, mv_tr_embedding = mv_reval.train(train_cor_dic)
            miscl_val, val_labels, mv_ts_embedding = mv_reval.test(test_cor_dic, modelfit)
            rndmisc_mean_val = mv_reval.rndlabels_traineval(mv_tr_embedding, mv_ts_embedding, tr_labels, val_labels)
            
            # If random labeling gives perfect prediction, substitute
            # misclassification for validation loop with 1.0
            if rndmisc_mean_val > 0.0:
                ms_val = miscl_val / rndmisc_mean_val
            else:
                ms_val = 1.0
            return len([n for n in np.unique(val_labels) if n >= 0]), miscl_tr, ms_val, tr_labels, val_labels
        except TypeError:
            return None, None, None, None, None

        
def _confint(vect):
    """
    Private function to compute confidence interval.

    :param vect: performance scores.
    :type vect: array-like
    :return: mean and error.
    :rtype: tuple
    """
    mean = np.mean(vect)
    # interval = 1.96 * math.sqrt((mean * (1 - mean)) / len(vect))
    interval = stats.t.ppf(1 - (0.05 / 2), len(vect) - 1) * (np.std(vect) / math.sqrt(len(vect)))
    return mean, interval



    # def evaluate_mcca(self, mod_tr, mod_ts, cov_tr, cov_ts, nclust=None, tr_lab=None):
    #     """
    #     Method that applies the selected clustering algorithm with the best number of clusters
    #     to the test set. It returns clustering labels.
    
    #     :param data_tr: training dataset.
    #     :type data_tr: ndarray, (n_samples, n_features)
    #     :param diagnosis_tr: diagnosis labels.
    #     :type diagnosis_tr: vector
    #     :param data_ts: test dataset.
    #     :type data_ts: ndarray, (n_samples, n_features)
    #     :param cov_tr: covariates for training set.
    #     :type cov_tr: ndarray, (n_samples, n_covariates)
    #     :param cov_ts: covariates for test set.
    #     :type cov_ts: ndarray, (n_samples, n_covariates)
    #     :param nclust: best number of clusters, default None.
    #     :type nclust: int
    #     :param tr_lab: clustering labels for the training set. If not None
    #         the clustering algorithm is not performed and the classifier is fitted.
    #         Available for clustering methods without `n_clusters` parameter. Default None.
    #     :type tr_lab: array-like
    #     :param combined_data: if True, only grey matter features will be corrected for TIV. 
    #         Otherwise, both grey matter and white matter features will be adjusted for the same set of covariates.
    #     :type: boolean value
    #     :return: labels and accuracy for both training and test sets.
    #     :rtype: namedtuple, (train_cllab: array, train_acc:float, test_cllab:array, test_acc:float)
    #     """
    #     data_tr_array = [np.array(mod_tr[mod]) for mod in mod_tr]
    #     data_ts_array = [np.array(mod_ts[mod]) for mod in mod_ts]
    #     cov_tr_array = [np.array(cov_tr[mod]) for mod in cov_tr]
    #     cov_ts_array = [np.array(cov_ts[mod]) for mod in cov_ts]
        
    #     if isinstance(tr_lab, list):
    #         tr_lab = np.array(tr_lab)
    #     try:
    #         if 'n_clusters' in self.clust_method.get_params().keys():
    #             self.clust_method.n_clusters = nclust
    #             train_cor_dic, test_cor_dic = super().GLMcorrection_multimodal(tr_idx, val_idx)
    #             tr_misc, modelfit, labels_tr = super().train(X_train_cor)
    #         elif 'n_components' in self.clust_method.get_params().keys():
    #             self.clust_method.n_clusters = nclust
    #             X_train_cor, X_test_cor = super().GLMcorrection(data_tr, diagnosis_tr, cov_tr, data_ts, cov_ts, num_col_DTI)
    #             tr_misc, modelfit, labels_tr = super().train(X_train_cor)
    #         else:
    #             X_train_cor, X_test_cor = super().GLMcorrection(data_tr, diagnosis_tr, cov_tr, data_ts, cov_ts, num_col_DTI)
    #             tr_misc, modelfit, labels_tr = super().train(X_train_cor, tr_lab)
    #         ts_misc, labels_ts = super().test(X_test_cor, modelfit)
    #     except TypeError:
    #         logging.info(f"Not possible to perform evaluation on the test set. "
    #                      f"No clusters were identified with "
    #                      f"{self.clust_method}")
    #         return None

    #     Eval = namedtuple('Eval',
    #                   ['train_cllab', 'train_acc', 'test_cllab', 'test_acc'])
    #     out = Eval(labels_tr, 1 - tr_misc, labels_ts, 1 - ts_misc)
    #     return out



"""
@author: Federica Colombo
         Psychiatry and Clinical Psychobiology Unit, Division of Neuroscience, 
         IRCCS San Raffaele Scientific Institute, Milan, Italy
"""

from matplotlib import pyplot as plt


def plot_metrics(cv_score,
                 figsize=(8, 5),
                 linewidth=1,
                 color=('black', 'black'),
                 legend_loc=2,
                 fontsize=12,
                 title="",
                 prob_lines=False,
                 save_fig=None):
    """
    Function that plots the average performance (i.e., normalized stability) over cross-validation
    for training and validation sets. The horizontal lines represent the random performance error
    for the correspondent number of clusters.

    :param cv_score: collection of cv scores as output by `reval.best_nclust_cv.FindBestCLustCV.best_nclust`.
    :type cv_score: dictionary
    :param figsize: (width, height), default (8, 5).
    :type figsize: tuple
    :param linewidth: width of the lines to draw.
    :type linewidth: int
    :param color: line colors for train and validation sets, default ('black', 'black').
    :type color: tuple
    :param legend_loc: legend location, default 2.
    :type legend_loc: int
    :param fontsize: size of fonts, default 12.
    :type fontsize: int
    :param title: figure title, default "".
    :type title: str
    :param prob_lines: plot the normalized stability of random labeling as thresholds, default False.
    :type prob_lines: bool
    :param save_fig: file name for saving figure in png format, default None.
    :type save_fig: str
    """
    fig, ax = plt.subplots(figsize=figsize)
    cl_list = list(cv_score['train'].keys())
    ax.plot(list(cv_score['train'].keys()),
            [me[0] for me in cv_score['train'].values()],
            linewidth=linewidth,
            linestyle='-.',
            label='training set',
            color=color[0])
    ax.errorbar(list(cv_score['val'].keys()),
                [me[0] for me in cv_score['val'].values()],
                [me[1][1] for me in cv_score['val'].values()],
                linewidth=linewidth,
                linestyle='-',
                label='validation set',
                color=color[1])
    if prob_lines:
        plt.hlines([(1 - (1 / k)) for k in cl_list], xmin=[k - 0.1 for k in cl_list],
                   xmax=[k + 0.1 for k in cl_list], linewidth=linewidth,
                   color=color[1])
    ax.legend(fontsize=fontsize, loc=legend_loc)
    plt.xticks([lab for lab in cv_score['train'].keys()], fontsize=fontsize)
    plt.yticks(fontsize=fontsize)
    plt.xlabel('Number of clusters', fontsize=fontsize)
    plt.ylabel('Normalized stability', fontsize=fontsize)
    plt.title(title)
    if save_fig is not None:
        plt.savefig(f'./{save_fig}', format=str.split(save_fig, '.')[1])
    else:
        plt.show()



# -*- coding: utf-8 -*-
"""
Created on Sat Oct 26 09:49:26 2024

@author: Federica Colombo
         Psychiatry and Clinical Psychobiology Unit, Division of Neuroscience, 
         IRCCS San Raffaele Scientific Institute, Milan, Italy
"""

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mvneureval",
    version="0.1.0",
    author="Federica Colombo",
    author_email="fcolombo.italia@gmail.com",
    description="Multiview stability-based relative clustering validation algorithm for multimodal data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fede-colombo/Multiview_NeuReval",
    download_url="https://github.com/fede-colombo/Multiview_NeuReval/releases/tag/v0.1.0",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["numpy",
                      "scipy",
                      "scikit-learn",
                      "umap-learn",
                      "matplotlib",
                      "mvlearn"],
    python_requires='>=3.6',
)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Federica Colombo
         Psychiatry and Clinical Psychobiology Unit, Division of Neuroscience, 
         IRCCS San Raffaele Scientific Institute, Milan, Italy
"""

import numpy as np
from sklearn.metrics import zero_one_loss
from mvneureval.utils import kuhn_munkres_algorithm
import logging
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier
import math

logging.basicConfig(format='%(asctime)s, %(levelname)s %(message)s',
                    datefmt='%H:%M:%S',
                    level=logging.INFO)


class RelativeValidationMultiview:
    """
    This class allows to perform the relative clustering validation procedure.
    A supervised algorithm is required to test cluster stability.
    Labels output from a clustering algorithm are used as true labels.

    :param s: initialized class for the supervised method.
    :type s: class
    :param c: initialized class for clustering algorithm.
    :type c: class
    :param preprocessing: initialized class for preprocessing method.
    :type preprocessing: class
    :param nrand: number of iterations to normalize cluster stability.
    :type nrand: int
    """

    def __init__(self, c, s, nrand=10):
        """
        Construct method.
        """
        self.clust_method = c
        self.class_method = s
        self.nrand = nrand
        
    def GLMcorrection_multimodal(self, modalities,covariates, tr_idx, val_idx):
        """
        Eliminate the confound of covariate, such as age and sex, from the input features.
        :param X_train: array, training features
        :param Y_train: array, training labels
        :param covar_train: array, training covariate data
        :param X_test: array, test labels
        :param covar_test: array, test covariate data
        :param num_col_DTI: index of dataset column corresponding to white matter features (i.e., "ACR" column)
        :return: corrected training & test feature data
        
        """
        X_train_dic = {mod:None for mod in modalities}
        X_test_dic = {mod:None for mod in modalities}
        cov_train_dic = {mod:None for mod in modalities}
        cov_test_dic = {mod:None for mod in modalities}
            
        for mod in modalities:
            X_train_dic[mod] = np.array(modalities[mod])[tr_idx]
            if X_train_dic[mod].ndim==1:
                X_train_dic[mod] = X_train_dic[mod].reshape(-1,1)
            X_test_dic[mod]= np.array(modalities[mod])[val_idx]
            if X_test_dic[mod].ndim==1:
                X_test_dic[mod] = X_test_dic[mod].reshape(-1,1)
            cov_train_dic[mod] = np.array(covariates[mod])[tr_idx]
            if cov_train_dic[mod].ndim==1:
                cov_train_dic[mod] = cov_train_dic[mod].reshape(-1,1)
            cov_test_dic[mod]= np.array(covariates[mod])[val_idx]
            if cov_test_dic[mod].ndim==1:
                cov_test_dic[mod] = cov_test_dic[mod].reshape(-1,1)
        
        
        # normalize the covariate z-scoring
        train_scaled_dic = {mod:None for mod in modalities}
        test_scaled_dic = {mod:None for mod in modalities}
        trcov_scaled_dic = {mod:None for mod in modalities}
        tscov_scaled_dic = {mod:None for mod in modalities}
        scaler = StandardScaler()
        
        for mod in modalities:
            train_scaled_dic[mod] = scaler.fit_transform(X_train_dic[mod])
            test_scaled_dic[mod] = scaler.fit_transform(X_test_dic[mod])
            trcov_scaled_dic[mod]= scaler.fit_transform(cov_train_dic[mod])
            tscov_scaled_dic[mod]= scaler.fit_transform(cov_test_dic[mod])
    
        # Adjust data for confounds of covariate
        train_cor_dic = {mod:None for mod in modalities}
        test_cor_dic = {mod:None for mod in modalities}
        
        for mod in modalities:
            beta = np.linalg.lstsq(trcov_scaled_dic[mod], train_scaled_dic[mod], rcond=None)
            train_cor_dic[mod] = (train_scaled_dic[mod].T - beta[0].T @ trcov_scaled_dic[mod].T).T
            test_cor_dic[mod] = (test_scaled_dic[mod].T - beta[0].T @ tscov_scaled_dic[mod].T).T
        
        return train_cor_dic, test_cor_dic

    def train(self, train_cor_dic, tr_lab=None):
        """
        Method that performs training. It compares the clustering labels on training set
        (i.e., A(X) computed by :class:`neureval.relative_validation_confounds.RelativeValidationConfounds.clust_method`) against
        the labels obtained from the classification algorithm
        (i.e., f(X), computed by :class:`neureval.relative_validation_confounds.RelativeValidationConfounds.class_method`).
        It returns the misclassification error, the supervised model fitted to the data,
        and both clustering and classification labels.

        :param X_train_cor: counfound corrected training dataset.
        :type X_train_cor: ndarray, (n_samples, n_features)
        :param tr_lab: cluster labels found during CV for clustering methods with no `n_clusters` parameter.
            If not None the clustering method is not performed on the whole test set. Default None.
        :type tr_lab: list
        :return: misclassification error, fitted supervised model object, clustering and classification labels.
        :rtype: float, object, ndarray (n_samples,)
        """
        mv_tr_data = [np.array(train_cor_dic[mod]) for mod in train_cor_dic]
        
        if tr_lab is None:
            clustlab_tr = self.clust_method.fit_predict(mv_tr_data)  # A_k(X)
        else:
            clustlab_tr = tr_lab
        if len([cl for cl in clustlab_tr if cl >= 0]) == 0:
            logging.info(f"No clusters found during training with {self.clust_method}.")
            return None
        
        mv_tr_embedding = self.clust_method.embedding_
        fitclass_tr = self.class_method.fit(mv_tr_embedding, clustlab_tr)
        classlab_tr = fitclass_tr.predict(mv_tr_embedding)
        misclass = zero_one_loss(clustlab_tr, classlab_tr)
        
        return misclass, fitclass_tr, clustlab_tr, mv_tr_embedding
        

    def test(self, test_cor_dic, fit_model):
        """
        Method that compares test set clustering labels (i.e., A(X'), computed by
        :class:`neureval.relative_validation_confounds.RelativeValidationConfounds.clust_method`) against
        the (permuted) labels obtained through the classification algorithm fitted to the training set
        (i.e., f(X'), computed by
        :class:`reval.relative_validation.RelativeValidationConfounds.class_method`).
        It returns the misclassification error, together with
        both clustering and classification labels.

        :param X_test_cor: confound corrected test dataset.
        :type X_test_cor: ndarray, (n_samples, n_features)
        :param fit_model: fitted supervised model.
        :type fit_model: class
        :param fit_preproc: fitted preprocessing method.
        :type fit_preproc: class
        :return: misclassification error, clustering and classification labels.
        :rtype: float, dictionary of ndarrays (n_samples,)
        """
        mv_ts_data = [np.array(test_cor_dic[mod]) for mod in test_cor_dic]
        
        clustlab_ts = self.clust_method.fit_predict(mv_ts_data)  # A_k(X')
        if len([cl for cl in clustlab_ts if cl >= 0]) == 0:
            logging.info(f"No clusters found during testing with {self.clust_method}")
            return None
        
        mv_ts_embedding = self.clust_method.embedding_
        classlab_ts = fit_model.predict(mv_ts_embedding)
        bestperm = kuhn_munkres_algorithm(np.int32(classlab_ts), np.int32(clustlab_ts))  # array of integers
        misclass = zero_one_loss(classlab_ts, bestperm)
        
        return misclass, bestperm, mv_ts_embedding

    def rndlabels_traineval(self, mv_tr_embedding, mv_ts_embedding, train_labels, test_labels):
        """
        Method that performs random labeling on the training set
        (N times according to
        :class:`neureval.relative_validation_confounds.RelativeValidationConfounds.nrand` instance attribute) and evaluates
        the fitted models on test set.

        :param X_train_cor: confound corrected training dataset.
        :type X_train_cor: ndarray, (n_samples, n_features)
        :param X_test_cor: confound corrected test dataset.
        :type X_test_cor: ndarray, (n_samples, n_features)
        :param train_labels: training set clustering labels.
        :type train_labels: ndarray, (n_samples,)
        :param test_labels: test set clustering labels.
        :type test_labels: ndarray, (n_samples,)
        :return: averaged misclassification error on the test set.
        :rtype: float
        """
        np.random.seed(0)
        shuf_tr = [np.random.permutation(train_labels)
                   for _ in range(self.nrand)]
        misclass_ts = list(map(lambda x: self._rescale_score_(mv_tr_embedding, mv_ts_embedding, x, test_labels), shuf_tr))
        return np.mean(misclass_ts)

    def _rescale_score_(self, xtr, xts, randlabtr, labts):
        """
        Private method that computes the misclassification error when predicting test labels
        with classification model fitted on training set with random labels.

        :param xtr: training dataset.
        :type xtr: ndarray, (n_samples, n_features)
        :param xts: test dataset.
        :type xts: ndarray, (n_samples, n_features)
        :param randlabtr: random labels.
        :type randlabtr: ndarray, (n_samples,)
        :param labts: test set labels.
        :type labts: ndarray, (n_samples,)
        :return: misclassification error.
        :rtype: float
        """
        self.class_method.fit(xtr, randlabtr)
        pred_lab = self.class_method.predict(xts)
        me_ts = zero_one_loss(pred_lab, kuhn_munkres_algorithm(np.int32(pred_lab), np.int32(labts)))
        return me_ts
